// Author: D. Rosseel
// Original: 27-07-2010
// Latest update: 05-08-2010
// Converted from Pascal to C by MickeyDelp.com

sbit PS2_Data_Mouse            at    RC0_bit;
sbit PS2_Clock_Mouse           at    RC1_bit;
sbit PS2_Data_Direction_Mouse  at TRISC0_bit;
sbit PS2_Clock_Direction_Mouse at TRISC1_bit;

#ifndef false
#define false 0
#endif

#ifndef true
#define true 1
#endif

typedef unsigned char byte;

typedef struct
{
 byte Keys;
 int  DeltaX;
 int  DeltaY;
 int  DeltaZ; // Scrollwheel
} TMouseData;

unsigned int PS2_Mouse_TimeOut;
unsigned int PS2_Mouse_TimeOut2;
byte PS2_Mouse_TimeOutError;
byte PS2_Mouse_ScrollWheel;

byte PS2_Mouse_Init();
// Initializes the Mouse Ps2 library && resets the mouse.
// Returns true if a "Mouse selftest passed" code is received from the Mouse, else false.
// Also enables the ScrollWheel (if any present).

byte PS2_Mouse_ReadData(TMouseData *MouseData);
// Reads the PS2 mouse data into "MouseData".
//   MouseData.Keys returns the pressed keys (one bit for every key)
//   MouseData.DeltaX returns the delta of the X value (hor position) with respect to the previous data
//   MouseData.DeltaY returns the delta of the Y value (vert position) with respect to the previous data
//   MouseData.DeltaZ returns the delta of the Scrollwheel value with respect to the previous data

void PS2_Mouse_Write(byte *Values, byte Count);
// Writes "Count" bytes out of "Values" to the PS2 Mouse. No reply expected here.

// The list of valid commands can be found in: http://www.computer-engineering.org/ps2mouse/

void PS2_Mouse_Write_Read(byte *Command, byte CountWrite, byte *Reply, byte CountRead);
// Same as "PS2_Mouse_Write" above, but after the write operation "CountRead" bytes are read from the mouse into "Reply".

byte PS2_Mouse_Reset();
// Resets the Mouse and waits (with a PS2_Mouse_TimeOut of 2 seconds) for the "selftest passed" acknowledge.
// Returns true if the "selftest passed" was acknowledged by the Mouse, otherwise false.
// Remark: the Mouse also resets itself (and does a selftest) at power up time (see "PS2_Mouse_Init").

byte PS2_Mouse_ScrollWheel_Present();
// Returns true if the mouse has a scrollwheel, else false.


byte WaitForClockLow()  // returns true if the clock went low within the PS2_Mouse_TimeOut
{
 unsigned int TmpW;
 TmpW = 0;
 do
 {
  TmpW++;
 } while ((PS2_Clock_Mouse != 0) && (TmpW < PS2_Mouse_TimeOut));
 return (PS2_Clock_Mouse == 0);
}

byte WaitForClockHigh()  // returns true if the clock went high within the PS2_Mouse_TimeOut
{
 unsigned int TmpW;
 TmpW = 0;
 do
 {
  TmpW++;
 } while ((PS2_Clock_Mouse != 1) && (TmpW < PS2_Mouse_TimeOut));
 return (PS2_Clock_Mouse == 1);
}

byte WaitForInitialClockLow()  // long timeout
{
 unsigned int TmpW;
 TmpW = 0;
 do
 {
  TmpW++;
 } while ((PS2_Clock_Mouse != 0) && (TmpW < PS2_Mouse_TimeOut2));
 return (PS2_Clock_Mouse == 0);
}

byte PS2_Mouse_Get_Byte()  // Reads one byte from the PS2 Mouse
{
 byte I, Tmp;

 PS2_Mouse_TimeOutError = false;
 if (!WaitForInitialClockLow())  // wait for first clockpulse with long timeout
 {
  PS2_Mouse_TimeOutError = true;
  return 0;
 }

 I = 0;
 Tmp = 0;
 while (I < 11)  // 11 clockpulses expected
 {
  if (!WaitForClockLow())  // wait for clock low (start of clockpulse)
  {
   PS2_Mouse_TimeOutError = true;
   return 0;
  }
  Delay_us(10);
  if ((I > 0) && (I < 9))
  {
   Tmp >>= 1;
   Tmp.B7 = PS2_Data_Mouse;
  }
  if (!WaitForClockHigh())  // wait for clock high (end of clockpulse)
  {
   PS2_Mouse_TimeOutError = true;
   return 0;
  }
  I++;
 }
 return Tmp;
}

byte PS2_Mouse_SelftestResult()
{
 byte Result = false;
 byte Tmp;
 unsigned int Cnt;

 // wait for selftest completion
 Cnt = 0;
 do
 {
  PS2_Clock_Mouse = 1;
  PS2_Clock_Direction_Mouse = 1;  // release clock (allow Mouse to send)
  Delay_us(2);

  Tmp = PS2_Mouse_Get_Byte();

  PS2_Clock_Mouse = 0;
  PS2_Clock_Direction_Mouse = 0;  // block clock (do not allow Mouse to send)
  Delay_us(2);

  if (!PS2_Mouse_TimeOutError)      // actual code received
  {
   if (Tmp == 0xAA)
    Result = true;        // reset completion code (selftest done)
   else if (Tmp == 0xFC)
    return false;         // error
  }
   
  Cnt++;
 } while ((Result != true) && (Cnt < 66)); // 66 times 30 msecs is approx 2 secs

  do                               // purge any extra bytes sent
  {
   PS2_Clock_Mouse = 1;
   PS2_Clock_Direction_Mouse = 1;  // release clock (allow Mouse to send)
   Delay_us(2);
   Tmp = PS2_Mouse_Get_Byte();
  } while (!PS2_Mouse_TimeOutError);

 PS2_Clock_Mouse = 0;
 PS2_Clock_Direction_Mouse = 0;  // block clock (do not allow Mouse to send)
 Delay_us(2);
 return Result;
}

byte PS2_Mouse_Enable_ScrollWheel()
{
 byte Cmd[7];
 byte Tmp;

 Cmd[0] = 0xF3;     // Set Sample rate command
 Cmd[1] = 200;      // 200
 Cmd[2] = 0xF3;
 Cmd[3] = 100;      // 100
 Cmd[4] = 0xF3;
 Cmd[5] = 80;       // 80
 Cmd[6] = 0xF2;     // get ID byte
 PS2_Mouse_Write(Cmd, 7);

 PS2_Clock_Mouse = 1;
 PS2_Clock_Direction_Mouse = 1;
 Delay_us(2);

 Tmp = PS2_Mouse_Get_Byte();

 PS2_Clock_Mouse = 0;
 PS2_Clock_Direction_Mouse = 0;
 Delay_us(2);

 return ((!PS2_Mouse_TimeOutError) && (Tmp == 3));
}

byte PS2_Mouse_ScrollWheel_Present()
{
  return PS2_Mouse_ScrollWheel;
}

byte PS2_Mouse_Reset()
{
  byte Arr;
  byte Result;

  PS2_Mouse_TimeOutError = false;
  PS2_Mouse_ScrollWheel  = false;
  Result = false;

  Arr = 0xFF;                          // reset command
  PS2_Mouse_Write(&Arr, 1);

  Result = PS2_Mouse_SelftestResult();

  PS2_Mouse_ScrollWheel = PS2_Mouse_Enable_ScrollWheel();

  Arr = 0xF4;                         // Enable Data Reporting
  PS2_Mouse_Write(&Arr, 1);

  PS2_Clock_Mouse = 0;                // force clock low (prevent mouse from sending data)
  PS2_Clock_Direction_Mouse = 0;
  Delay_us(2);                        // prevent read after write problem
  return Result;
}

byte PS2_Mouse_Init()  // Initializes the Mouse Ps2 routines
{
  PS2_Mouse_TimeOutError = false;
  PS2_Mouse_ScrollWheel = false;

  PS2_Mouse_TimeOut  = (unsigned int)(20 * Clock_MHz());  // gives approx 1 ms of timeout
  PS2_Mouse_TimeOut2 = 30 * PS2_Mouse_TimeOut;            // approx 30 ms

  PS2_Clock_Mouse = 0;            // force clock low
  PS2_Clock_Direction_Mouse = 0;
  Delay_us(2);                    // prevent read after write problem

  PS2_Data_Mouse = 1;
  PS2_Data_Direction_Mouse = 1;
  Delay_us(2);
  
  PS2_Mouse_SelftestResult();
  return PS2_Mouse_Reset();       // give extra reset
}

byte PS2_Mouse_ReadData(TMouseData *MouseData)  // Reads 3 or 4 bytes from the Ps2 Mouse into "MouseData"
{
 byte Index, Tmp;
 byte TmpCodes[4] = {0,0,0,0};
 byte MaxCount;

 if (PS2_Mouse_ScrollWheel)
  MaxCount = 4;
 else
  MaxCount = 3;

  PS2_Mouse_TimeOutError = false;

 PS2_Clock_Mouse = 1;
 PS2_Clock_Direction_Mouse = 1;  // release clock (allow Mouse to send)
 Delay_us(2);

/*
 // This logic from orignal Pascal code was removed because it causes missed events.
 // wait for first low clock with (short) timeout (low clock does not occur when no key is pressed: nothing to send)
 if (!WaitForClockLow())
 {
  PS2_Clock_Mouse = 0;  
  PS2_Clock_Direction_Mouse = 0;    // block clock
  Delay_us(2);
  return 0;                         // no new data
 }
*/

 Index = 0;
 do  // get bytes until all bytes read
 {
  Tmp = PS2_Mouse_Get_Byte();
  if (PS2_Mouse_TimeOutError)
  {
    PS2_Clock_Mouse = 0;
    PS2_Clock_Direction_Mouse = 0;  // block clock
    Delay_us(2);
    return 0;                       // Read Error
  }
  TmpCodes[Index] = Tmp;
  Index++;
 }
 while (Index != MaxCount);

 PS2_Clock_Mouse = 0;
 PS2_Clock_Direction_Mouse = 0;     // block clock (prevent Mouse to send)
 Delay_us(2);

 MouseData->Keys = TmpCodes[0] & 7;

 if (TmpCodes[0].B4 == 1)           // X "sign" bit
  MouseData->DeltaX = TmpCodes[1] | 0xFF00;
 else
  MouseData->DeltaX = TmpCodes[1];

 if (TmpCodes[0].B5 == 1)          // Y "sign" bit
  MouseData->DeltaY = TmpCodes[2] | 0xFF00;
 else
  MouseData->DeltaY = TmpCodes[2];

 if (TmpCodes[3].B7 == 1)          // scrollwheel sign bit  
  MouseData->DeltaZ = TmpCodes[3] | 0xFF00;
 else
  MouseData->DeltaZ = TmpCodes[3];

 return true;
}

byte WaitForClockHighThenLow()
{
 unsigned int TmpW;
 TmpW = 0;
 do
 {
  TmpW++;
 }
 while ((PS2_Clock_Mouse != 1) && (TmpW < PS2_Mouse_TimeOut));

 if (PS2_Clock_Mouse != 1)
  return false;

 TmpW = 0;
 do
 {
  TmpW++;
 }
 while ((PS2_Clock_Mouse != 0) && (TmpW < PS2_Mouse_TimeOut));

 return (PS2_Clock_Mouse == 0);
}

byte WaitForDataLowThenHigh()
{
 unsigned int TmpW;
 TmpW = 0;
 do
 {
  TmpW++;
 }
 while ((PS2_Data_Mouse != 0) && (TmpW < PS2_Mouse_TimeOut));

 if (PS2_Data_Mouse != 0)
  return false;

 TmpW = 0;
 do
 {
  TmpW++;
 }
 while ((PS2_Data_Mouse != 1) && (TmpW < PS2_Mouse_TimeOut));

 return (PS2_Data_Mouse == 1);
}


void PS2_Mouse_Write_Byte(byte Value)  // writes 1 byte to the PS2 Mouse (command or data)
{
 byte Parity, I;
 PS2_Mouse_TimeOutError = false;

 // calculate parity
 Parity = 0;
 for (I = 1; I; I<<=1)
  if (!(Value & I))
   Parity++;
 if (Parity.B0 == 1) // odd
  Parity = 0;
 else
  Parity = 1;

  // make a request to send condition
  PS2_Clock_Mouse = 0;  
  PS2_Clock_Direction_Mouse = 0;  // block clock
  Delay_us(200);
  PS2_Data_Mouse = 0;  
  PS2_Data_Direction_Mouse = 0;   // Make data line zero
  Delay_us(2);
  PS2_Clock_Direction_Mouse = 1; 
  PS2_Clock_Mouse = 1;            // release clock line
  Delay_us(2);

  if (!WaitForInitialClockLow())
  {
    PS2_Mouse_TimeOutError = true;
    goto ExitWriteByte;
  }

  // send data and parity
  for (I = 0; I <= 8; I++)
  {
   if (I < 8)
   {
    PS2_Data_Mouse = Value.B0;   // send LSB bit
    Value >>= 1;                 // get next bit
   }
   else
    PS2_Data_Mouse = Parity;     // send parity

  if (!WaitForClockHighThenLow())
  {
   PS2_Mouse_TimeOutError = true;
   goto ExitWriteByte;
  }
 }

 PS2_Data_Mouse = 1;
 PS2_Data_Direction_Mouse = 1;   // release dataline

 if (!WaitForClockHighThenLow())
 {
  PS2_Mouse_TimeOutError = true;
  goto ExitWriteByte;
 }

 if (!WaitForDataLowThenHigh())  // Check acknowledge (data line)
 {
  PS2_Mouse_TimeOutError = true;
  goto ExitWriteByte;
 }

 ExitWriteByte:
 // done, release data && clock
 PS2_Data_Direction_Mouse  = 1;
 PS2_Data_Mouse  = 1;
 PS2_Clock_Direction_Mouse = 1;
 PS2_Clock_Mouse = 1;
}

void PS2_Mouse_Write(byte *Values, byte Count)  // Writes "Count" bytes out of "Values" to the PS2 Mouse.
{
 byte Index, Reply;
 Index = 0;
 while (Count > 0)
 {
  PS2_Mouse_Write_Byte(Values[Index]);  // send byte to the Mouse
  if (PS2_Mouse_TimeOutError)
   break;

  Reply = PS2_Mouse_Get_Byte();         // get the reply (0xFA is Ok)
  if (PS2_Mouse_TimeOutError)
   break;

  Index++;
  Count--;
 }

 PS2_Clock_Mouse = 0;  
 PS2_Clock_Direction_Mouse = 0;   // block clock
 Delay_us(2);
 PS2_Data_Direction_Mouse  = 1; 
 PS2_Data_Mouse   = 1;           // release data line
 Delay_us(2);
}

void PS2_Mouse_Write_Read(byte *Command, byte CountWrite, byte *Reply, byte CountRead)
{
 byte I;
 PS2_Mouse_Write(Command, CountWrite);
 if (CountRead > 0)
 {
  I = 0;
  PS2_Clock_Mouse = 1;
  PS2_Clock_Direction_Mouse = 1;  // enable the mouse to send
  Delay_us(2);
  while (I < CountRead)
  {
   Reply[I] = PS2_Mouse_Get_Byte();
   I++;
  }
  PS2_Clock_Mouse = 0;  
  PS2_Clock_Direction_Mouse = 0;  // disable the mouse to send
  Delay_us(2);
 }
}